/*
   piktc.c -- PIKT master control program, and config file compiler

   $Header: /home/cvs/ucgsb/pikt/src/piktc/piktc.c,v 1.65 2000/02/24 02:09:45 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA

   --------------------------------------------------------------------------

   Usage:  piktc <-cCdefGhikKlrRstTvxX>
                 [+|-D          <define(s)>]
                 [+C            <command(s)>]
                 [+|-A|S[f] all|<alert(s)/script(s)>]
                 [+|-P[f]   all|<program(s)>]
                 [+|-F[f]   all|<file(s)>]
                 [+|-O[f]   all|<object(s)>]
                  +|-H      all|<host(s)>
                 [ALL]

           -c                         syntax check all config files
           -C                         syntax doublecheck all config files
           -d                         disable alert(s)
           -e                         enable alert(s)
           -f                         diff alert/script/program/file/object
                                        file(s)
           -G                         run in debug mode
           -h                         show program help
           -i                         install alert/script/program/file/object
                                        file(s)
           -k                         kill alert daemon (piktd)
           -K                         kill service daemon (pikt_svc)
           -l                         list alert/script/program/file/object
                                        file(s)
           -r                         (re)start alert daemon (piktd)
           -R                         (re)start service daemon (piktc_svc)
           -s                         show alert(s) status
           -t                         delete alert/script/program/file/object/
                                        history/log file(s)
           -T                         run in test mode
           -v                         run in verbose mode
           -x                         execute alert(s)/script(s)
           -X                         execute alert(s)/script(s) with no wait
           +|-D          <define(s)>  define/undefine define(s)
           +C            <command(s)> include command string(s)
           +|-A|S[f] all|<alert(s)>   include/exclude (fix) alert(s)/script(s)
           +|-P[f]   all|<program(s)> include/exclude (fix) program file(s)
           +|-F[f]   all|<file(s)>    include/exclude (fix) other file(s)
           +|-O[f]   all|<object(s)>  include/exclude (fix) object file(s)
           +|-H      all|<host(s)>    include/exclude host(s)/os(es)/group(s)/
                                        alias(es)
           ALL                        +A all +S all +P all +F all +O all +H all

 */

#include "../lib/common.h"
#include "piktc.h"

#include <rpc/rpc.h>
#include "rpc.h"

const char piktc_version[] = "PIKT piktc version " VERSION;

/* command-line options */
int ischeck = FALSE;
int isdebug = FALSE;
int isdelete = FALSE;
int isdiff = FALSE;
int isdoublecheck = FALSE;
int isdisable = FALSE;
int isenable = FALSE;
int isexec = FALSE;
int isexecwait = FALSE;
int ishelp = FALSE;
int isinstall = FALSE;
int iskill_piktd = FALSE;
int iskill_piktc_svc = FALSE;
int islist = FALSE;
int isrestart_piktd = FALSE;
int isrestart_piktc_svc = FALSE;
int istest = FALSE;
int isshow = FALSE;
int isverbose = FALSE;

int bailout = FALSE;	/* signifies command-line error,
			   or error in systems.cfg
			 */

int prescan;
int postscan;

/* PIKT.conf options */
int uid = 0;		/* default to root */
char *master = (char *) NULL;
char *domain = (char *) NULL;
char *accessmode = (char *) NULL;
char *master_address = (char *) NULL;
int isauth_by_master = TRUE;
int isauth_by_master_address = FALSE;
int rpc_timeout = RPC_TIMEOUT;
int rpc_retry_timeout = RPC_RETRY_TIMEOUT;
int rpcinfo_retries = RPCINFO_RETRIES;
int isverboselog = FALSE;
int isautosyslog = FALSE;
int istcponly = FALSE;
int execcmdout_limit = EXECCMDOUT_LIMIT;

char hostname[MAXIDENT+1];
char domainname[MAXIDENT+1];

char **argvcurr;	/* current command-line argument */
char **argvlast;	/* last command-line argument */

const char piktnullchar[]    =       { '\0' };
const char piktspacechar[]   = { ' ',  '\0' };
const char pikttabchar[]     = { '\t', '\0' };
const char piktnewlinechar[] = { '\n', '\0' };

unsigned loffset = 0;

int cfgerr = FALSE;
int stzerr = FALSE;

unsigned long cfbs = 0L;	/* initially, no config files detected yet */
unsigned long rcfbs = 1L;	/* only systems.cfg is required */
unsigned long mcfbs = 0L;	/* initially, no config files are considered
				   missing
				 */

size_t hbslongs;
size_t dbslongs;

int hostnum = 0;
int opsysnum = 0;
int defnum = 0;

char **hosts;
char *hostcurr;

char **opsys;

unsigned long *hbscurr = (unsigned long *) NULL;

/* cmdcommands/cmdalerts/cmdprograms/cmdfiles/cmdobjects/cmdhosts are the
   commands/alerts/programs/files/objectsets/hosts specified in the command line

   cfgalerts/cfgprograms/cfgfiles/cfgobjects are the
   alerts/programs/files/objectsets included in the tagged config files

   commands/alerts/programs/files/objects are the
   commands/alerts/programs/files/objectsets being written (or executed)
   for the current host in the current program run

   since we bailout if any alert/program/file/objectsset (or host) specified
   in the command line is bogus (meaning it's not found in the config files),
   cmdalerts/cmdprograms/cmdfiles/cmdobjects is a subset of
   cfgalerts/cfgprograms/cfgfiles/cfgobjects

   commands/alerts/programs/files/objects is identical to
   cmdcommands/cmdalerts/cmdprograms/cmdfiles/cmdobjects if comand-line
   commands/alerts/programs/files/objects are specified, else
   cmdcommands/cmdalerts/cmdprograms/cmdfiles/cmdobjects is empty and
   commands/alerts/programs/files/objects are the things actually written
   (or executed)
 */

nam *commands,               *incmdcommands,                 *cmdcommands,
    *alerts,   *cfgalerts,   *incmdalerts,   *excmdalerts,   *cmdalerts,
    *programs, *cfgprograms, *incmdprograms, *excmdprograms, *cmdprograms,
    *files,    *cfgfiles,    *incmdfiles,    *excmdfiles,    *cmdfiles,
    *objects,  *cfgobjects,  *incmdobjects,  *excmdobjects,  *cmdobjects,
               *cfghosts,    *incmdhosts,    *excmdhosts,    *cmdhosts;

int commandsall, alertsall, programsall,	/* was "all" specified */
    filesall, objectsall, hostsall;		/* on commandline?     */

FILE	*stripin,	*stripout,
	*tagin,		*tagout,
	*sysin,		*sysout,
	*cfgin,		*cfgout;

char *infile = (char *) NULL;
int fd[2];
pid_t pid;
int lineno;

/* the order of the following .cfg files is vitally important! */
const char *cfgfilenames[] = { "systems.cfg",
                               "defines.cfg",
                               "macros.cfg",
                               "alarms.cfg",
                               "alerts.cfg",
                               "programs.cfg",
                               "files.cfg",
                               "objects.cfg"
                             };

char *cfgfileprev = (char *) NULL;
char *cfgfilecurr = (char *) NULL;

int cfgcurr;
int cfglast;

cfgln *configs[OBJCFG+1];	/* assumes OBJCFG is last cfg file */

/* the order of the following is vitally important! */
const char *cfgbof[] = {        "/\n",	/* SYSCFG */
                               "//\n",	/* DEFCFG */
                              "///\n",	/* MCRCFG */
                             "////\n",	/* ALMCFG */
                            "/////\n",	/* ALTCFG */
                           "//////\n",	/* PRGCFG */
                          "///////\n",	/* FILCFG */
                         "////////\n"	/* OBJCFG */
                       };

char *cfgline;

cfgln *clplast = (cfgln *) NULL;

unsigned bof;
unsigned bon;
unsigned bol;

char *mins, *hrs, *doms, *mons, *dows;
char *drift;
char *priority;
char *nicecmd, *mailcmd, *lpcmd;

int stripalm, stripprg, stripfil, stripobj;

int hasmacro;
int domacros;
int dostrip;

int incnum;

char piktcprog[13];			/* PIKTCPROG string */
char piktcvers[5];			/* PIKTCVERS string; PIKTCVERS <= 99 */

unsigned long services = ALL_SERVICES;	/* all services enabled by default */

int waitpid_options = 0
#ifdef NEED_WAITPID_WNOHANG
                        | WNOHANG
#endif
                        ;

void preprocess(int, int, int);
void processcmd(int, char **);
void process(char *, unsigned long*, int, char **);
void mkcmds(char *);
void dodoublecheck(void);
int missingcfgfiles(int);

int
main (int argc, char *argv[])
{
	static char pidstr[8];

	init(argc, argv, (void (*)(void))piktc_usage);
	if (bailout)
		exit(1);

	if (! isfile(CPCMD, TRUE) || ! isfile(DIFFCMD, TRUE) || ! isfile(RMCMD, TRUE) || ! isfile(EGREPCMD, TRUE) || ! isfile(RPCINFOCMD, TRUE)) {
		/* error(s) logged in PIKTLOG */
		exit(1);
	}

	/* initializations for rpcinfo probes */
	sprintf(piktcprog, " %ld ", PIKTCPROG);
	sprintf(piktcvers, " %d ", PIKTCVERS);

	commands = incmdcommands = cmdcommands =
	alerts = cfgalerts = incmdalerts = excmdalerts = cmdalerts =
	programs = cfgprograms = incmdprograms = excmdprograms = cmdprograms =
	files = cfgfiles = incmdfiles = excmdfiles = cmdfiles =
	objects = cfgobjects = incmdobjects = excmdobjects = cmdobjects =
	          cfghosts = incmdhosts = excmdhosts = cmdhosts = (nam *)NULL;
	stripalm = stripprg = stripfil = stripobj = TRUE;

	argvcurr = argv+1;	/* skip the program name */
	argvlast = argv+argc-1;

	postscan = FALSE;
	cmdparse();		/* further parse the command line */
	if (bailout) {
		piktc_usage();
		exit(1);
	}

	/* apply the lock, to prevent other concurrent piktc sessions */
	if (islock(piktclock)) {
		fprintf(stderr, "%s found!  Another piktc session may be in progress.\nTry again later (else rm the file).\n", piktclock);
		fflush(stderr);
		exit(1);
	} else {
		pid = getpid();
		sprintf(pidstr, "%u", (unsigned)pid);
		if (mklock(piktclock, pidstr, TRUE, TRUE) == ERROR)
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in main(), mklock() error for %s", piktclock);
	}

	/* preprocess the config files; the order is vitally important! */

	/* always preprocess systems.cfg */
	preprocess(SYSCFG, TRUE, TRUE);
	preprocess(SYSCFG, TRUE, FALSE);

	/* we can't tell from the command line if preprocessing the
	   defines.cfg file is necessary, since defines might be set and
	   referred to later on in other config files, so we always
	   preprocess defines.cfg (possibly just a faked "dummy" file)
	 */
	preprocess(DEFCFG, TRUE, TRUE);
	preprocess(DEFCFG, TRUE, FALSE);

	/* if we have 'piktc -x|X +C "<commands without macros>" +H ...',
	   we don't need to process any more config files
	 */
	if ((isexec || isexecwait) && ! ((rcfbs & (1L << MCRCFG)) || (rcfbs & (1L << ALTCFG)))) {
		;	/* no further preprocessing required */
	} else {
		preprocess(MCRCFG, TRUE, FALSE);
		if (missingcfgfiles(MCRCFG)) {
			bailout = TRUE;
		/* if we have 'piktc -x|X +C "<commands with macros>" +H ...',
		   we don't need to process any more config files
		 */
		} else if ((isexec || isexecwait) && (rcfbs < (1L << ALMCFG))) {
			;	/* no further preprocessing required */
		} else {
			/* if we have 'piktc -c|C ...' */
			/* or if we have 'piktc ... +A|S ...' */
			if (ischeck || isdoublecheck || rcfbs & (1L << ALMCFG))
				preprocess(ALMCFG, stripalm, FALSE);
			/* or if we have 'piktc ... +A|S ...' */
			if (ischeck || isdoublecheck || rcfbs & (1L << ALTCFG))
				preprocess(ALTCFG, TRUE, FALSE);
			/* or if we have 'piktc ... +P ...' */
			if (ischeck || isdoublecheck || rcfbs & (1L << PRGCFG))
				preprocess(PRGCFG, stripprg, FALSE);
			/* or if we have 'piktc ... +F ...' */
			if (ischeck || isdoublecheck || rcfbs & (1L << FILCFG))
				preprocess(FILCFG, stripfil, FALSE);
			/* or if we have 'piktc ... +O ...' */
			if (ischeck || isdoublecheck || rcfbs & (1L << OBJCFG))
				preprocess(OBJCFG, stripobj, FALSE);
			if (missingcfgfiles(OBJCFG))
				bailout = TRUE;
		}
	}

	if (! bailout)
		preprocesscmd();	/* preprocess command-line items */
	
	if (! bailout)
		processcmd(argc, argv);		/* execute the command(s) */

	if (bailout)
		piktc_usage();

	/* remove the lock */
	if (rmlocks(TRUE, TRUE) == ERROR)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in main(), rmlocks() error");

	if (bailout)
		exit(1);
	else
		exit(0);
}

/* output program usage to screen
 */
void
piktc_usage (void)
{
	printf("\nUsage:  piktc <-cCdefGhikKlrRstTvxX>\n\t      [+|-D          <define(s)>]\n\t      [+C            <command(s)>]\n\t      [+|-A|S[f] all|<alert(s)/script(s)>]\n\t      [+|-P[f]   all|<program(s)>]\n\t      [+|-F[f]   all|<file(s)>]\n\t      [+|-O[f]   all|<object(s)>]\n\t       +|-H      all|<host(s)>\n\t      [ALL]\n\n");
	printf("\t-c\t\t\t   syntax check all config files\n");
	printf("\t-C\t\t\t   syntax doublecheck all config files\n");
	printf("\t-d\t\t\t   disable alert(s)\n");
	printf("\t-e\t\t\t   enable alert(s)\n");
	printf("\t-f\t\t\t   diff alert/script/program/file/object\n\t\t\t\t     file(s)\n");
	printf("\t-G\t\t\t   run in debug mode\n");
	printf("\t-h\t\t\t   show program help\n");
	printf("\t-i\t\t\t   install alert/script/program/file/object\n\t\t\t\t     file(s)\n");
	printf("\t-k\t\t\t   kill alert daemon (piktd)\n");
	printf("\t-K\t\t\t   kill service daemon (piktc_svc)\n");
	printf("\t-l\t\t\t   list alert/script/program/file/object\n\t\t\t\t     file(s)\n");
	printf("\t-r\t\t\t   (re)start alert daemon (piktd)\n");
	printf("\t-R\t\t\t   (re)start service daemon (piktc_svc)\n");
	printf("\t-s\t\t\t   show alert(s) status\n");
	printf("\t-t\t\t\t   delete alert/script/program/file/object\n\t\t\t\t     /history/log file(s)\n");
	printf("\t-T\t\t\t   run in test mode\n");
	printf("\t-v\t\t\t   run in verbose mode\n");
	printf("\t-x\t\t\t   execute alert(s)/script(s)\n");
	printf("\t-X\t\t\t   execute alert(s)/script(s) with no wait\n");
	printf("\t+|-D          <define(s)>  define/undefine define(s)\n");
	printf("\t+C            <command(s)> include command string(s)\n");
	printf("\t+|-A|S[f] all|<alert(s)>   include/exclude (fix) alert(s)\n");
	printf("\t+|-P[f]   all|<program(s)> include/exclude (fix) program file(s)\n");
	printf("\t+|-F[f]   all|<file(s)>    include/exclude (fix) other file(s)\n");
	printf("\t+|-O[f]   all|<object(s)>  include/exclude (fix) object file(s)\n");
	printf("\t+|-H      all|<host(s)>    include/exclude host(s)/os(es)/group(s)/\n\t\t\t\t     alias(es)\n");
	printf("\tALL\t\t\t   +A all +S all +P all +F all +O all +H all\n\n");
	return;
}

/* for each host in the host-bit set, process the piktc command: execute
   commands/alarms/scripts, enable/disable alerts, install/delete files,
   list files, kill/restart alert daemon, show alert status, check config
   file validity
   note: "NIL" is a dummy extension for signalling unambiguously that a file
   is a program from programs.cfg, not from files.cfg
 */
void
processcmd (int argc, char **argv)
{
	ident *idp;
	nam *hstp;
	char cmd[MAXCMD+1];
	static unsigned long *hbs;
	static unsigned long *tbs;
	unsigned long hbsl;
	register int i, j;

	/* for each host in the command-line set, turn on its corresponding
	   bit in the host bitset
	 */
	hbs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), "processcmd");
	foreach (hstp, cmdhosts) {
		idp = lookupident(hstp->name, SYSCFG);
		if (idp) {
			tbs = hbs;
			hbs = bsunion(hbs, idp->bs, hbslongs);
			free(tbs);
		} else	/* validatecmditems() should have caught this */
			fprintf(stderr, "unknown host or group:  %s\n", hstp->name);
			fflush(stderr);
	}
	/* foreach unsigned long in the host bitset */
	for (i=0; i<hbslongs; i++) {
		hbsl = *(hbs+i);
		/* foreach bit in the current unsigned long */
		for (j=0; j<8*sizeof(unsigned long); j++) {
			if (isdebug)
				logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "i is %d, j is %d, 1<<j is %x, hbsl is %x", i, j, 1<<j, hbsl);
			/* if the bit for this host is on */
			if (hbsl & (1<<j)) {
				hostcurr = *(hosts+((i*sizeof(unsigned long)*8)+j));
				FREE(hbscurr);
				hbscurr = bsdup(getbs(hostcurr, HST, FALSE), hbslongs);
				cfgerr = FALSE;
				stzerr = FALSE;

				if (isverbose) {
					printf("\n%s%sing %s...\n", isdoublecheck ? "double" : "", (ischeck || isdoublecheck) ? "check" : "process", hostcurr);
					fflush(stdout);
				}

				/* if not doing a syntax check, determine if
				   remote piktc_svc is running; if not, say
				   so, and bypass the current host
				 */
				if (! (ischeck || isdoublecheck)) {
					if (chkpiktc_svc(hostcurr) != 0) {
						if (isverbose) {
							fprintf(stderr, "remote piktc_svc not running\n");
							fflush(stderr);
						}
						continue;
					}
				}

				/* disable remote alerts */
				if ((/* isinstall || */ isdisable || isenable || isdelete) && cmdalerts)
					disable_alerts(hostcurr, alertsall ? cfgalerts : cmdalerts, isverbose && ! isenable /* && ! isinstall */);

				/* delete remote alert files */
				if ((/* isinstall || */ isdelete) && cmdalerts)
					delete_files(hostcurr, alertsall ? cfgalerts : cmdalerts, altdir, ".alt", isverbose && ! isinstall);

				/* delete remote history/log files */
				if (isdelete && cmdalerts) {
					delete_files(hostcurr, alertsall ? cfgalerts : cmdalerts, hstdir, ".hst", isverbose && ! isinstall);
					delete_files(hostcurr, alertsall ? cfgalerts : cmdalerts, logdir, ".log", isverbose && ! isinstall);
				}

				/* delete remote object files */
				if ((/* isinstall || */ isdelete) && cmdobjects)
					delete_files(hostcurr, objectsall ? cfgobjects : cmdobjects, objdir, ".obj", isverbose && ! isinstall);

				/* install/enable/execute/check/
				   doublecheck/diff cfg/dat files */
				if (isinstall || isenable || isexec || isexecwait || ischeck || isdoublecheck || isdiff || (islist && (cmdprograms || cmdfiles)) || (isdelete && (cmdprograms || cmdfiles))) {

					/* clear out diffing dir */
					if (isdiff) {
						sprintf(cmd, "%s -rf %s 2>/dev/null", RMCMD, difdir);
						system(cmd);
						/* allow time for just-
						   attempted directory clearing
						 */
						sleep(TIME_TO_SLEEP);
						if (mkdir(difdir, (mode_t)(S_IRWXU|S_IRGRP|S_IWGRP)) == -1)	/* 750 */
							logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in processcmd(), mkdir() failure for %s", difdir);
					}

					/* clear out staging dir */
					sprintf(cmd, "%s -rf %s 2>/dev/null", RMCMD, stgdir);
					system(cmd);
					/* allow time for just-
					   attempted directory clearing
					 */
					sleep(TIME_TO_SLEEP);
					if (mkdir(stgdir, (mode_t)(S_IRWXU|S_IRGRP|S_IWGRP)) == -1)	/* 750 */
						logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in processcmd(), mkdir() failure for %s", stgdir);

					/* process config files for current host */
					process(hostcurr, hbscurr, argc, argv);

					/* garbage collect */
					cleanup();

					if (isdoublecheck) {
						dodoublecheck();
					}

					if (cfgerr || ischeck || isdoublecheck)
						continue;

				}

				/* note: it is necessary to delete program
				   and other files after invoking process(),
				   for we need to know the programs and other
				   files list for the current machine;
				   otherwise, for programs and other files
				   installed outside the default PIKT program
				   directory and inside a directory NFS
				   cross-mounted to more than one machine,
				   we risk blowing them away by passing
				   cfgprograms or cfgfiles to delete_files()
				   (in the manner we do above) when processing
				   a later machine
				 */

				/* delete remote program files */
				if ((/* isinstall || */ isdelete) && cmdprograms)
					delete_files(hostcurr, programs, prgdir, "NIL", isverbose && ! isinstall);

				/* delete remote other files */
				if ((/* isinstall || */ isdelete) && cmdfiles)
					delete_files(hostcurr, files, "", "", isverbose && ! isinstall);

				/* note: an alert/program/file/object may be
				   specified in the command line but not be
				   written to the staging directory, because
				   it doesn't apply to a particular host,
				   which is why we can't fetch/install/enable
				   cmdalerts/cmdprograms/cmdfiles/cmdobjects,
				   which might include
				   alerts/programs/files/objects not processed
				 */

				/* install remote alert files */
				if (isinstall && cmdalerts)
					install_files(hostcurr, alerts, altdir, ".alt", MODE_FIL_DEFAULT, UID_DEFAULT, GID_DEFAULT, isverbose);

				/* install remote program files */
				if (isinstall && cmdprograms)
					install_files(hostcurr, programs, prgdir, "NIL", MODE_PRG_DEFAULT, UID_DEFAULT, GID_DEFAULT, isverbose);

				/* install remote other files */
				if (isinstall && cmdfiles)
					install_files(hostcurr, files, "", "", MODE_FIL_DEFAULT, UID_DEFAULT, GID_DEFAULT, isverbose);

				/* install remote object files */
				if (isinstall && cmdobjects)
					install_files(hostcurr, objects, objdir, ".obj", MODE_FIL_DEFAULT, UID_DEFAULT, GID_DEFAULT, isverbose);

				/* fetch remote alert files */
				if (isdiff && cmdalerts)
					fetch_files(hostcurr, alerts, altdir, ".alt", isverbose);

				/* fetch remote program files */
				if (isdiff && cmdprograms)
					fetch_files(hostcurr, programs, prgdir, "NIL", isverbose);

				/* fetch remote other files */
				if (isdiff && cmdfiles)
					fetch_files(hostcurr, files, "", "", isverbose);

				/* fetch remote object files */
				if (isdiff && cmdobjects)
					fetch_files(hostcurr, objects, objdir, ".obj", isverbose);

				/* diff the diffing & staging dirs */
				if (isdiff) {
					if (isverbose) {
						printf("diffing file(s)...\n");
						fflush(stdout);
					}
					sprintf(cmd, "%s -r %s %s | %s -iv \"common subdir\"", DIFFCMD, stgdir, difdir, EGREPCMD);
					system(cmd);
				}

				/* enable remote alerts */
				if ((/* isinstall || */ isenable) && cmdalerts)
					enable_alerts(hostcurr, alertsall ? alerts : alerts, isverbose);

				/* execute remote commands */
				if ((isexec || isexecwait) && cmdcommands)
					execute_commands(hostcurr, commandsall ? commands : commands, isverbose);

				/* execute remote alerts */
				if ((isexec || isexecwait) && cmdalerts)
					execute_alerts(hostcurr, alertsall ? alerts : alerts, isverbose);

				/* list remote alert files */
				if (islist && cmdalerts)
					list_files(hostcurr, alertsall ? cfgalerts: cmdalerts, altdir, ".alt", isverbose);

				/* list remote program files */
				if (islist && cmdprograms)
					list_files(hostcurr, programsall ? cfgprograms : cmdprograms, prgdir, "NIL", isverbose);

				/* list remote other files */
				if (islist && cmdfiles)
					list_files(hostcurr, filesall ? cfgfiles : cmdfiles, "", "", isverbose);

				/* list remote object files */
				if (islist && cmdobjects)
					list_files(hostcurr, objectsall ? cfgobjects : cmdobjects, objdir, ".obj", isverbose);

				/* kill remote alert daemon (piktd) */
				if (iskill_piktd || isrestart_piktd)
					kill_daemon(hostcurr, KILL_PIKTD_SERVICE, isverbose && ! isrestart_piktd);

				/* kill remote service daemon (piktc_svc) */
				if (iskill_piktc_svc)
					kill_daemon(hostcurr, KILL_PIKTC_SVC_SERVICE, isverbose && ! isrestart_piktc_svc);

				/* (re)start remote alert daemon (piktd) */
				if (isrestart_piktd) {
					/* allow time for daemon to die from
					   just-concluded kill
					 */
					sleep(2*TIME_TO_SLEEP);
					restart_daemon(hostcurr, RESTART_PIKTD_SERVICE, isverbose);
				}

				/* (re)start remote service daemon (piktc_svc) */
				if (isrestart_piktc_svc) {
					/* allow time for daemon to die from
					   just-concluded kill
					 */
					sleep(2*TIME_TO_SLEEP);
					restart_daemon(hostcurr, RESTART_PIKTC_SVC_SERVICE, isverbose);
				}

				/* show remote alert status */
				if (isshow && cmdalerts)
					show_alerts_stata(hostcurr, alertsall ? cfgalerts : cmdalerts, isverbose);

				retire(&commands);
				retire(&alerts);
				retire(&programs);
				retire(&files);
				retire(&objects);
			}
		}
	}
	register_request("", "", FALSE);	/* in effect, deregister; this
						   is need here especially in
						   the isexec case, but should
						   be redundant for everything
						   else
						 */
	free(hbs);
	return;
}

/* open the config file, filter out C++ style comments in strip, and
   pass the output to the sys parser for systems.cfg, and to the
   cfg parser for all other .cfg files; after preprocessing, we no
   longer need refer to config files on disk, as all information is
   stored in internal data structures
 */
void
preprocess (int cfgtype, int strip, int isprescan)
{
	static char _preprocess_[] = "preprocess";
	struct stat fbuf;
	unsigned long *hbs;
	int status;

	/* check for existence of cfg file; in certain cases, bailout if it
	   doesn't exist
	 */
	/* linux chokes on this
	FREE(cfgfileprev);
	 */
	cfgfileprev = (char *) NULL;
	/* linux chokes on this
	FREE(cfgfilecurr);
	 */
	cfgfilecurr = infile = (char *) mymalloc((strlen(cfgdir)+strlen(cfgfilenames[cfgtype])+2) * sizeof(char), _preprocess_);
	sprintf(infile, "%s/%s", cfgdir, cfgfilenames[cfgtype]);
	if (stat(infile, &fbuf) != 0) {
		switch (cfgtype) {
			/* systems.cfg will always be mandatory
			 */
			case SYSCFG:
				logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in preprocess(), stat() error for %s", infile);
				break;
			case DEFCFG:
				/* even if no actual defines.cfg, fake one so
				   as to avoid a possible reduce/reduce
				   error in def.y that might result if all
				   config files, except systems.cfg, are
				   optional
				 */
				if (! isprescan) {
					cfgcurr = cfgtype;
					hbs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), _preprocess_);
					fill(hbs, hbslongs);
					tag(1, cfgfilecurr, hbs, hbslongs, FALSE, FALSE, "dummy FALSE\n");
					FREE(hbs);
					cfbs |= (1L << cfgtype);
					lineno = 1;
					logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, TRUE, "in preprocess(), stat() error for %s", infile);
				}
				return;
				break;
			case MCRCFG:
			case ALMCFG:
			case ALTCFG:
			case PRGCFG:
			case FILCFG:
			case OBJCFG:
				logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, TRUE, "in preprocess(), stat() error for %s", infile);
				return;
				break;
		}
	} else {
		if (cfgtype > SYSCFG)
			cfbs |= (1L << cfgtype);	/* add this cfg file
							   to config bitset
						 	 */
		else
			cfbs = 1L;
	}
	cfgcurr = cfgtype;
	dostrip = strip;
	prescan = isprescan;
	if (! isprescan) {
		switch (cfgtype) {
			case SYSCFG:
				hbslongs = (int)((hostnum-1)/(8*sizeof(long))) + 1;
				hosts = (char **) mymalloc(hostnum*(sizeof(char *)), _preprocess_);
				opsys = (char **) mymalloc(opsysnum*(sizeof(char *)), _preprocess_);
				hostnum = 0;
				opsysnum = 0;
				hbsfull = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), _preprocess_);
				hbsempty = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), _preprocess_);
				break;
			case DEFCFG:
				dbslongs = (int)((defnum-1)/(8*sizeof(long))) + 1;
				gdbs = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), _preprocess_);
				gubs = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), _preprocess_);
				/* currently serve no purpose
				dbsfull = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), _preprocess_);
				dbsempty = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), _preprocess_);
				 */
				break;
			default:
				break;
		}
	}
	if (! isprescan)
		domacros = TRUE;
	incnum = 0;
	lineno = 1;
	if (dostrip) {
		if (pipe(fd) < 0)
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in preprocess(), pipe() error");
		/* open the ipc channels */
		ipcopen(incnum);
		switch (pid = fork()) {
			case -1:	/* error */
				logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in preprocess(), fork() error");
				break;
			/* pipe from child to parent */
			case 0:		/* child */
				close(fd[0]);
				stripin = fopen(infile, "r");
				stripout = fdopen(fd[1], "w");
				wait_parent(incnum);
				tell_parent(incnum);
				striplex();
				fclose(stripin);
				fclose(stripout);
				close(fd[1]);
				tell_parent(incnum);
				wait_parent(incnum);
				_exit(0);
				break;
			default:	/* parent */
				close(fd[1]);
				switch (cfgtype) {
					case SYSCFG:
						sysin = fdopen(fd[0], "r");
						tell_child(incnum);
						wait_child(incnum);
						sysparse();
						fclose(sysin);
						break;
					default:
						tagin = fdopen(fd[0], "r");
						tell_child(incnum);
						wait_child(incnum);
						tagparse();
						fclose(tagin);
						break;
				}
				close(fd[0]);
				wait_child(incnum);
				tell_child(incnum);
				waitpid(pid, &status, waitpid_options);
				break;
		}
		/* close the ipc channels */
		ipcclose(incnum);
	} else {
		/* for now, only ALMCFG/PRGCFG/FILCFG/OBJCFG are potentially
		   not stripped, SYSCFG is always stripped, so switch
		   not really necessary here
		 */
		switch (cfgtype) {
			case SYSCFG:
				sysin = fopen(infile, "r");
				sysparse();
				fclose(sysin);
				break;
			default:
				tagin = fopen(infile, "r");
				tagparse();
				fclose(tagin);
				break;
		}
	}
	return;
}

/* foreach macros, alarms, alerts, programs, files & objects config file,
   process the config lines; foreach config line, if its tag matches for
   the current host, then input it to the cfg parser
 */
void
process (char *host, unsigned long *hbs, int argc, char **argv)
{
	int i;
	int lastcfg;
	struct stat fbuf;

	#ifdef DEEPDEBUG
	if (isdebug)
		dumpcfgs(hbs);
	#endif

	/* determine the last config file to be processed (the first is
	   always sytems.cfg)
	 */
	lastcfg = 0;	/* systemsc.fg (SYSCFG = 0) always required */
	for (i=1; i<=OBJCFG; i++) {
		if (cfbs & (1L << i))
			lastcfg = i;
	}

	/* only call defparse() if there is a defines.cfg file
	 */
	/* linux chokes on this
	FREE(cfgfileprev);
	 */
	cfgfileprev = (char *) NULL;
	/* linux chokes on this
	FREE(cfgfilecurr);
	 */
	cfgfilecurr = infile = (char *) mymalloc((strlen(cfgdir)+strlen(cfgfilenames[DEFCFG])+2) * sizeof(char), "process");
	sprintf(infile, "%s/%s", cfgdir, cfgfilenames[DEFCFG]);
	if (stat(infile, &fbuf) == 0) {
		empty(gdbs, dbslongs);
		empty(gubs, dbslongs);
		defnum = 0;
		loffset = 0;
		cfgcurr = DEFCFG;
		cfglast = DEFCFG;
		bof = TRUE;
		defparse();
		if (bailout) {	/* set in deferror() */
			if (rmlocks(TRUE, TRUE) == ERROR)
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in process(), rmlocks() error");
			exit(1);
		}
		/* now go back and apply any command-line def overrides */
		argvcurr = argv+1;	/* skip the program name */
		argvlast = argv+argc-1;
		postscan = TRUE;
		cmdparse();
		if (bailout) {		/* the first scan should have
					   caught any errors
					 */
			piktc_usage();
			exit(1);
		}
		loffset = 0;
		cfgcurr = MCRCFG;
		cfglast = lastcfg;
		bof = TRUE;
		defparse();
		#ifdef DEEPDEBUG
		/* only dump a second time after the lines are marked
		   for doproc
		 */
		if (isdebug)
			dumpcfgs(hbs);
		#endif
		if (bailout) {	/* set in deferror() */
			if (rmlocks(TRUE, TRUE) == ERROR)
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in process(), rmlocks() error");
			exit(1);
		}
	}

	loffset = 0;
	cfgcurr = MCRCFG;
	cfglast = lastcfg;
	cfgfileprev = (char *) NULL;
	cfgfilecurr = (char *) cfgfilenames[MCRCFG];
	bof = TRUE;
	bon = FALSE;
	bol = FALSE;
	makepiktmacros();
	cfgparse();
	mkcmds(host);
	return;
}

/* for every cmdcommands (if any), macro expand the command string, and
   assign to the commands nam list
 */
void
mkcmds (char *host)
{
	static char _mkcmds_[] = "mkcmds";
	nam *nmp1, *nmp2, *nnmp;
	char *expcmd;

	foreach (nmp1, cmdcommands) {
		expcmd = expand(nmp1->name);
		nnmp = (nam *) mymalloc(sizeof(nam), _mkcmds_);
		nnmp->next = (nam *) NULL;
		nnmp->name = mystrdup(nmp1->name, _mkcmds_);
		nnmp->aux = expcmd;
		nnmp->dir = (char *) NULL;
		nnmp->file = (char *) NULL;
		nnmp->mode = (char *) NULL;
		nnmp->uid = (char *) NULL;
		nnmp->gid = (char *) NULL;
		nnmp->state = IGNORE;
		/* add to commands list */
		if (commands) {
			for (nmp2=commands; nmp2->next; nmp2=nmp2->next) ;
			nmp2->next = nnmp;
		} else
			commands = nnmp;
	}
	return;
}

/* do a second syntax check on each alert in STGDIR, by running them through
   a pikt -C check (which does a more thorough syntax check than piktc);
   if problems are discovered in the pikt doublecheck, it might be nice to
   save each specimen for later inspection, but we'll reserve implementation
   of that for another time!
 */
void
dodoublecheck (void)
{
	nam *nmp;
	static char *alt;
	char cmd[MAXCMD+1];
	struct stat fbuf;

	foreach (nmp, alerts) {
		if (nmp->name) {
			FREE(alt);
			alt = (char *) mymalloc((strlen(stgdir)+strlen(nmp->name)+6) * sizeof(char), "dodoublecheck");
			sprintf(alt, "%s/%s.alt", stgdir, nmp->name);
			if (lstat(alt, &fbuf) == 0) {	/* alert file found */
				sprintf(cmd, "%s/pikt -C +A %s", bindir, nmp->name);
				system(cmd);
			}
		}
	}
	FREE(alt);
	return;
}

/* check to see if piktc_svc on host is up; return 0 if so, ERROR if down;
   in the case of 'timed out' replies, we reattempt a number of times
 */
int
chkpiktc_svc (char *host)
{
	char cmd[MAXCMD+1];
	char line[MAXLINE+1];
	FILE *rpcinfo;
	char *cp;
	int try;

	/* for some mysterious reason, in AIX only, a successful popen()
	   call below randomly returns empty output; this is an aftereffect
	   of the IPC piping in preprocess(); until we solve this mystery,
	   we just automatically report successful contact here, but at
	   the risk of crashing piktc later on when and if it attempts to
	   do RPC with a downed system, or an up system but with no running
	   piktc_svc
	 */
	#ifdef AIX
	return 0;	/* automatic success in AIX case */
	#endif

	/* note: before, we had been using 'rpcinfo -t ...' successfully,
	   until irix balked (timed out with no response); so we switched
	   to 'rpcinfo -u ...', which seems to make everyone happy
	 */
	sprintf(cmd, "%s -u %s %lu | %s '\n' ' ' 2>&1", RPCINFOCMD, host, (unsigned long) RPCPRGNUM, "tr" /* TRCMD */);
	try = 0;
	line[0] = '\0';
	while (++try <= 1+rpcinfo_retries) {
		if ((rpcinfo = popen(cmd, "r")) == NULL) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in chkpiktc_svc(), popen() error");
			return ERROR;
		}
		if (isverboselog || isdebug)
			logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "probing rpcinfo on %s", host);
		while (fgets(line, MAXLINE+1, rpcinfo) != NULL) {
			if ((cp = strchr(line, '\n')) != NULL)
				*cp = '\0';
			if (isverboselog || isdebug)
				logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "%s", line);

			if (strstr(line, piktcprog) && strstr(line, piktcvers)) {
				pclose(rpcinfo);
				return 0;
			}
		}
		pclose(rpcinfo);
	}
	if ((cp = strchr(line, '\n')) != NULL)
		*cp = '\0';
	logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in chkpiktc_svc(), for remote host %s: %s", host, line[0] ? line : "unknown rpcinfo failure");
	return ERROR;
}

/* check to see if any required config files are missing */
int
missingcfgfiles (int lastcfg)
{
	register int i;
	unsigned long tcfbs;

	tcfbs = 0L;
	for (i=0; i<=lastcfg; i++)
		tcfbs |= (1L << i);
	tcfbs &= rcfbs;
	if ((mcfbs = (cfbs & tcfbs) ^ tcfbs)) {
		for (i=0; i<=lastcfg; i++) {
			if (! (mcfbs ^ (1L << i)))
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in missingcfgfiles(), missing %s file", cfgfilenames[i]);
		}
		return TRUE;
	}
	return FALSE;
}
