/*
   log.c -- PIKT pikt & piktd log routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/log.c,v 1.23 2000/02/13 05:03:58 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"
#include "piktd.h"

const char *leveltypes[] = { "EMERG", "ALERT", "CRIT", "ERR", "WARNING", "NOTICE", "INFO", "DEBUG" };

int syslog_facility = LOG_LOCAL0;

static void logit(int, int, const char *, int, int, const char *, va_list ap);

/* nonfatal error; log the message, then return to the caller
 */
void
logreturn (int logfiletype, int logtype, const char *srcfile, int lno, int syscall, const char *fmt, ...)
{
	va_list ap;

	#ifdef PIKTD
	if (/* ! logdir || */ ! piktdlog) {
	#else
	if (/* ! logdir || */ ! piktlog) {
	#endif
		if ((logfiletype != SYSLOG) || ! isautosyslog) {
			return;
		}
	}
	va_start(ap, fmt);
	logit(logfiletype, logtype, srcfile, lno, syscall, fmt, ap);
	va_end(ap);
	return;
}

/* fatal error; log the message, then exit the program
 */
void
logexit (int logfiletype, int logtype, const char *srcfile, int lno, int syscall, const char *fmt, ...)
{
	va_list ap;

	#ifdef PIKTD
	if (/* ! logdir || */ ! piktdlog) {
	#else
	if (/* ! logdir || */ ! piktlog) {
	#endif
		if ((logfiletype != SYSLOG) || ! isautosyslog) {
			#ifdef PIKTD
			rmlocks(FALSE, FALSE);
			#else
			rmlocks(interactive, FALSE);
			#endif
			exit(2);
		}
	}
	va_start(ap, fmt);
	logit(logfiletype, logtype, srcfile, lno, syscall, fmt, ap);
	va_end(ap);
	#ifdef PIKTD
	rmlocks(FALSE, FALSE);
	#else
	rmlocks(interactive, FALSE);
	#endif
	exit(2);
}

/* format the message, adding a date/time stamp, a label, source file &
   source line number, if a serious error, and system error message,
   as appropriate; print the message, either to a file or to stderr
 */
static void
logit (int logfiletype, int logtype, const char *srcfile, int lno, int syscall, const char *fmt, va_list ap)
{
	int errno_save;
	char buf[MAXBUF+1];	/* arbitrarily large */
	struct tm *tm;
	time_t tnow;
	FILE *logfile;

	errno_save = errno;
	tnow = time((time_t*)0);
	tm = localtime(&tnow);
	strcpy(buf, asctime(tm));
	switch (logtype) {
		case LOG_ERR:
			sprintf(buf+20, "ERROR: (%s, line %d) ", srcfile, lno);
			break;
		default:
			sprintf(buf+20, "%s: ", leveltypes[logtype]);
			break;
	}
	vsprintf(buf+strlen(buf), fmt, ap);
	#if defined(SHOW_SYSCALL)
	if (syscall)
		sprintf(buf+strlen(buf), ": %s", strerror(errno_save));
	#endif
	strcat(buf, "\n");
	switch (logfiletype) {
		case PIKTLOG:
			if (ischeck || isdoublecheck) {
				fflush(stdout);
				fputs(buf+4, stderr);
				fflush(stderr);
			} else {
				#ifdef PIKTD
				if ((logfile = fopen(piktdlog, "a")) == NULL) {
					fflush(stdout);
					fprintf(stderr, "ERROR: in logit(), fopen() error for %s\n", piktdlog);
					fflush(stderr);
					rmlocks(FALSE, FALSE);
					exit(1);
				}
				fputs(buf+4, logfile);
				if (fclose(logfile)) {
					fflush(stdout);
					fprintf(stderr, "ERROR: in logit(), fclose() error for %s\n", piktdlog);
					fflush(stderr);
					rmlocks(FALSE, FALSE);
					exit(1);
				}
				chmod(piktdlog, (mode_t)(S_IRUSR|S_IWUSR));  /* 600 */
				#else
				if (isdebug) {
					fflush(stdout);
					fputs(buf+4, stderr);
					fflush(stderr);
				} else {
					if ((logfile = fopen(piktlog, "a")) == NULL) {
						fflush(stdout);
						fprintf(stderr, "ERROR: in logit(), fopen() error for %s\n", piktlog);
						fflush(stderr);
						rmlocks(interactive, FALSE);
						exit(1);
					}
					fputs(buf+4, logfile);
					/* if interactive, send error
					   message also to stderr
					 */
					if (interactive && ((logtype == LOG_ERR) /* || (logtype == WARNING) */ )) {
						fflush(stdout);
						fputs(buf+4, stderr);
						fflush(stderr);
					}
					if (fclose(logfile)) {
						fflush(stdout);
						fprintf(stderr, "ERROR: in logit(), fclose() error for %s\n", piktlog);
						fflush(stderr);
						rmlocks(interactive, FALSE);
						exit(1);
					}
					chmod(piktlog, (mode_t)(S_IRUSR|S_IWUSR));  /* 600 */
				}
				#endif
			}
			break;
		#ifndef PIKTD
		case LOCLOG:
			if (isdebug) {
				fflush(stdout);
				fputs(buf+4, stderr);
				fflush(stderr);
			} else {
				if ((logfile = fopen(stmcurr->logfile, "a")) == NULL) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in logit(), fopen() error for %s", stmcurr->logfile);
					return;
				}
				fputs(buf+4, logfile);
				if (fclose(logfile)) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in logit(), fclose() error for %s", stmcurr->logfile);
					return;
				}
				chmod(stmcurr->logfile, (mode_t)(S_IRUSR|S_IWUSR)); /* 600 */
			}
			break;
		case SYSLOG:
			if (isdebug) {
				fflush(stdout);
				fputs(buf+4, stderr);
				fflush(stderr);
			} else {
				if (! isautosyslog) {
					openlog("pikt", LOG_CONS|LOG_PID, syslog_facility);
					syslog(LEVEL, "%s", buf+20);
					closelog();
				}
			}
			break;
		default:
			break;
		#endif
	}
	if (isautosyslog) {
		#ifdef PIKTD
		openlog("piktd", LOG_CONS|LOG_PID, syslog_facility);
		#else
		openlog("pikt", LOG_CONS|LOG_PID, syslog_facility);
		#endif
		syslog(logtype, "%s", buf+20);
		closelog();
	}
	return;
}
